// Experimental algorithm to detect accidental byte swapping in 16-bit linear PCM
// Some results: anywhere from 17 to 26 dB difference in amplitude between straight and byteswapped versions of
// the Brandenberg Concertos, distributed as FLAC files


#define _DARWIN_FEATURE_64_BIT_INODE 1 // For OSX
#define FILE_OFFSET_BITS 64 // For Linux (and BSD?)

#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <locale.h>
#include <math.h>
#include <unistd.h>
#include "dspfec.h"

#define PAGESIZE 4096

int main(int argc,char *argv[]){
  int c,i,fd,j;
  char *filename;
  unsigned char *p;
  signed short *sample;
  long long energy1,energy2;
  signed short *swap_buffer;
  struct stat statbuf;
  size_t size;
  char *locale;
  double decibels;
  
  // Set defaults
  size = 2048;
  locale = "en_US.UTF-8";

  // Process command-line args
  while((c = getopt(argc,argv,"")) != EOF){
  }
  cpu_id();

  if(setlocale(LC_ALL, locale) == NULL){
    fprintf(stderr, "setlocale(%s) failed\n", locale);
  }
  swap_buffer = malloc(size * sizeof(*swap_buffer));
  if(swap_buffer == NULL){
    printf("malloc of swap buffer with size %ld failed\n",(long)size);
    exit(1);
  }
  for(i=optind;i<argc;i++){
    filename = argv[i];

    if(stat(filename,&statbuf) != 0){
      printf("%s: can't stat '%s': %s; skipping\n",argv[0],filename,strerror(errno));
      continue;
    }
    if(!S_ISREG(statbuf.st_mode)){
      printf("%s: '%s' is not a regular file; skipping\n",argv[0],filename);
      continue;
    }
    if((fd = open(filename,O_RDONLY)) == -1){
      printf("%s: can't open '%s': %s; skipping\n",argv[0],filename,strerror(errno));
      continue;
    }
    p = NULL;
    p = mmap(&p,statbuf.st_size,PROT_READ,MAP_FILE|MAP_SHARED,fd,(off_t)0);
    if(p == MAP_FAILED){
      printf("%s: file '%s' memory mapping failed: %s; skipping\n",argv[0],filename,strerror(errno));
      close(fd);
      continue;
    }
    sample = (signed short *)(p + 4096); // Skip first page with file headers, etc
    size = (statbuf.st_size - 4096)/sizeof(signed short);
    energy1 = sumsq(sample,size);

    energy2 = 0;
    size = PAGESIZE/sizeof(signed short);
    while(sample + size < (signed short *)&p[statbuf.st_size]){
      for(j=0;j<size;j++)
	swap_buffer[j] = ((sample[j] >> 8) & 0xff) | ((sample[j] << 8) & 0xff00);
      energy2 += sumsq(swap_buffer,size);
      sample += size;
    }
    decibels = 10 * log10((double)energy1/(double)energy2);
    printf("%s: file '%s' straight energy %'llu byteswapped energy %'llu difference %.1lf dB\n",argv[0],filename,energy1,energy2,decibels);
    munmap(p,statbuf.st_size);
    close(fd);
  }
  exit(0);
}
