// Determine CPU features for vector arithmetic, word size

#include <stdio.h>
#include "dspfec.h"

int Verbose;
int Cpu_features;
char *Cpu_feature_names[] = {
  "UNKNOWN", "NONE", "ALTIVEC", "MMX", "SSE", "SSE2", "SSE3", "SSSE3", "SSE41", "SSE42"
};

#if defined(__ppc__) && defined(__VEC__)
#include <sys/sysctl.h>
#endif

void compilation_type(){
  fprintf(stderr,"Compiled for ");

#if __x86_64__
  fprintf(stderr,"x86-64");
#elif __i686__
  fprintf(stderr,"i686");
#elif __i586__
  fprintf(stderr,"i586");
#elif __i486__
  fprintf(stderr,"i486");
#elif __i386__
  fprintf(stderr,"i386");
#elif __ppc__
  fprintf(stderr,"ppc");
#else
  fprintf(stderr,"unknown CPU");
#endif

  fprintf(stderr," with");

#if __SSE4_2__
  fprintf(stderr," SSE4.2");
#endif
#if __SSE4_1__
  fprintf(stderr," SSE4.1");
#endif
#if __SSE4A__
  fprintf(stderr," SSE4a");
#endif
#if __SSE4__
  fprintf(stderr," SSE4");
#endif
#if __SSSE3__
  fprintf(stderr," SSSE3");
#endif
#if __SSE3__
  fprintf(stderr," SSE3");
#endif
#if __SSE2__
  fprintf(stderr," SSE2");
#endif
#if __SSE__
  fprintf(stderr," SSE");
#endif
#if __MMX__
  fprintf(stderr," MMX");
#endif
#if __3DNOW__
  fprintf(stderr," 3DNOW");
#endif

#if __ALTIVEC__
  fprintf(stderr," Altivec/VMX/Velocity Engine");
#endif

  fprintf(stderr,"\n");
}


// Determine vector features of CPU, if any, and display along with compiler options and word size
void cpu_id(){
  
#if 0
  if(Verbose)
    compilation_type();
#endif

  if(Cpu_features == UNKNOWN){
#ifdef __x86_64__
    {
      int _eax,_ebx,_ecx,_edx;
      int op;
      
      op = 1;
      
      // Got this from an IBM tutorial on the web. It even works for both 32-bit and
      // 64-bit modes!
      asm("cpuid" : "=a" (_eax), "=b" (_ebx), "=c" (_ecx), "=d" (_edx) : "a" (op));
      
      if(_ecx & (1<<20)){
	Cpu_features = SSE42;
      } else if(_ecx & (1<<19)) {
	Cpu_features = SSE41;
      } else if(_ecx & (1<<9)) {
	Cpu_features = SSSE3;
      } else if(_ecx & (1<<0)) {
	Cpu_features = SSE3;
      } else if(_edx & (1<<26)) {
	Cpu_features = SSE2;
      } else if(_edx & (1<<25)) {
	Cpu_features = SSE;
      } else if(_edx & (1<<23)) {
	Cpu_features = MMX;
      } else {
	Cpu_features = NONE;    
      }
    }
#elif (__i386__ || __i486__ || __i586__ || __i686__)
    {
      int regs[4];
      
      regs[0] = 1;
      asm ("mov %0,%%edi; mov %%ebx,%%esi; mov  0(%%edi),%%eax; mov  4(%%edi),%%ebx; mov  8(%%edi),%%ecx; mov 12(%%edi),%%edx; \
        cpuid;								\
        mov %%eax, 0(%%edi); mov %%ebx, 4(%%edi); mov %%ecx,8(%%edi); mov %%edx,12(%%edi); mov %%esi,%%ebx"
	   : : "g"(regs) : "%eax","%ecx","%edx","%edi","%esi");
      
      if(regs[2] & (1<<20)){
	Cpu_features = SSE42;
      } else if(regs[2] & (1<<19)) {
	Cpu_features = SSE41;
      } else if(regs[2] & (1<<9)) {
	Cpu_features = SSSE3;
      } else if(regs[2] & (1<<0)) {
	Cpu_features = SSE3;
      } else if(regs[3] & (1<<26)) {
	Cpu_features = SSE2;
      } else if(regs[3] & (1<<25)) {
	Cpu_features = SSE;
      } else if(regs[3] & (1<<23)) {
	Cpu_features = MMX;
      } else {
	Cpu_features = NONE;    
      }
    }
#elif defined(__ppc__) && defined(__VEC__)
    {
      /* Ask the OS if we have Altivec support */
      int selectors[2] = { CTL_HW, HW_VECTORUNIT };
      int hasVectorUnit = 0;
      size_t length = sizeof(hasVectorUnit);
      int error = sysctl(selectors, 2, &hasVectorUnit, &length, NULL, 0);
      if(0 == error && hasVectorUnit)
	Cpu_features = ALTIVEC;
      else
	Cpu_features = NONE;
    }
#else
  Cpu_features = NONE;
#endif
  } // Cpu_features == UNKNOWN
  if(Verbose)
    fprintf(stderr,"Address size: %lu bits; vector features: %s\n",(unsigned long)8*sizeof(void *),Cpu_feature_names[Cpu_features]);
}

