/* 16-bit signed integer dot product
 * Use MMX - NOT DONE YET!
 *
 * Copyright Oct 2009 Phil Karn
 * May be used under the terms of the GNU public license
 *
 * $Id$
 */
#include <stdlib.h>
#include <assert.h>
#include "sse.h"

#if !defined(__MMX__)
#error "Must compile with -mmmx or better"
#endif

/* Internal definitions for dotproduct function */
struct dotprod_mmx {
  int len; /* Number of coefficients */

  /* On MMX, these hold 4 copies of the coefficients,
   * preshifted by 0,1,...3 words to meet all possible input data
   * alignments (see Intel ap559 on MMX dot products).
   *
   */
  v4hi *coeffs[4];
};


/* Create and return a descriptor for use with the dot product function */
void *initdp_mmx(signed short coeffs[],int len){
  struct dotprod_mmx *dp;
  int j,i;
  signed short *wp;

  if(len <= 0)
    return NULL;

  dp = (struct dotprod_mmx *)calloc(1,sizeof(struct dotprod_mmx));
  dp->len = len;

  wp = (signed short *)malloc(sizeof(signed short)*(len+3)*4);
  assert(((long)wp & 0x7) == 0);
  /* Make 4 copies of coefficients, one for each data alignment */
  for(i=0;i<4;i++){
    dp->coeffs[i] = (v4hi *)wp;
    for(j=0;j<i;j++)
      *wp++ = 0; /* Leading zeroes; wp += i */
    for(j=0;j<len;j++)
      *wp++ = coeffs[j];
    while((long)wp & 0x7)
      *wp++ = 0; /* Trailing zeroes; wp += 3 - (3+len+i) % 4 */
  }
  assert(wp <= (signed short *)dp->coeffs[0] + (len+3)*4); /* Ensure we haven't overflowed */
  return (void *)dp;
}

/* Free a dot product descriptor created earlier */
void freedp_mmx(void *p){
  struct dotprod_mmx *dp = (struct dotprod_mmx *)p;

  free(dp->coeffs[0]);
  free(dp);
}

/* Compute a dot product given a descriptor and an input array
 * The length is taken from the descriptor
 */

/* MMX-assisted version */
int dotprod_mmx(void *p,signed short a[]){
  struct dotprod_mmx *dp = (struct dotprod_mmx *)p;
  int al,i,vcnt;
  v4hi *ar,*coeffs;
  union {
    int d[2]; /* Must be 32 bits */
    v2si v;
  } sums;
  v2si s;

  /* Round down to 8-byte boundary. We'll operate
   * on data before the user's buffer, but the corresponding
   * coefficients are zero so there's no harm. A segfault can't occur
   * either because the page boundaries are much larger than 16 bytes
   */
  ar = (v4hi *)((long)a & ~7);
  al = (short *)a - (short *)ar; /* Number of words misaligned */
  
  vcnt = (dp->len+al+3)/4; /* Number of vectors, rounding up to include partial ones */
  coeffs = dp->coeffs[al];
  
  /* Do 4 signed 16-bit multiply-adds per loop. Each pair of multiply-adds is combined into
   * one of two 32-bit sums.
   */
  s = (v2si)_mm_madd_pi16(coeffs[0],ar[0]);          /* [A2*B2+A3*B3]    [A0*B0+A1*B1] */
  for(i=1;i<vcnt;i++){
    s = _mm_add_pi32(_mm_madd_pi16(coeffs[i],ar[i]),s);
  }
  /* return the sum of the two 32-bit sub sums */
  sums.v = s;
  _mm_empty();
 return sums.d[0] + sums.d[1];
}

/* Temporarily use C code for 64-bit version; doing this in MMX may be difficult */
long long dotprodl_mmx(void *p,signed short a[]){
  struct dotprod_mmx *dp = (struct dotprod_mmx *)p;
  long long corr = 0;
  int i;

  for(i=0;i<dp->len;i++){
    corr += (long)(a[i]) * ((signed short *)dp->coeffs[0]) [i];
  }
  return corr;
}
