#if !defined(__SSE4_1__)
#error "Must compile with -msse4.1 or better"
#endif

#include "sse.h"
#include "dotprod_sse2.h"

long long dotprodl_sse41(void *p,signed short a[]){
  struct dotprod_sse2 *dp = (struct dotprod_sse2 *)p;
  int al,i,vcnt;
  v8hi *ar,*coeffs;
  v2di sums; /* two 64-bit sums */
  union {
    long long s[2];
    v2di v;
  } s;

  /* Round down to 16-byte boundary. We'll operate
   * on data before the user's buffer, but the corresponding
   * coefficients are zero so there's no harm. A segfault can't occur
   * either because the page boundaries are much larger than 16 bytes
   */
  ar = (v8hi *)((long)a & ~15);
  al = (short *)a - (short *)ar; /* Number of words misaligned */
  
  vcnt = (dp->len+al+7)/8; /* Number of vectors, rounding up to include partial ones */
  sums = _mm_setzero_si128();
  coeffs = dp->coeffs[al];
  
  /* Do 8 signed 16-bit multiply-adds per loop. Each pair of multiply-adds is combined into
   * one of four 32-bit sums. This is just wide enough for one pair, so to prevent overflow
   * we immediately extend each to 64 bits and add into two 64-bit running sub-sums.
   */
  for(i=0;i<vcnt;i++){
    v4si isums;
    isums = _mm_madd_epi16(coeffs[i],ar[i]);                   /* [A6*B6+A7*B7]   [A4*B4+A5*B5]  [A2*B2+A3*B3]    [A0*B0+A1*B1] */

    /* SSE4.1 has a 32-64-bit sign extend, but SSE2 does not.
     * But could this actually be slower??
     */
    sums += _mm_cvtepi32_epi64(isums);                         /* Sum low two 32-bit sums into the 64-bit final running sum */
    sums += _mm_cvtepi32_epi64(_mm_srli_si128(isums,8));       /* And sum the high two 32-bit sums */
  }
  /* return the sum of the two 64-bit sub sums */
  sums += _mm_unpackhi_epi64(sums,_mm_setzero_si128());
  s.v = sums;
  return s.s[0];
}
