#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <limits.h>
#include <assert.h>
#include <getopt.h>
#include <math.h>
#include <locale.h>
#include "dspfec.h"


int main(int argc,char *argv[]){

  signed short *data;
  unsigned long long ssq, blkno, total_samples;
  unsigned long samples;
  long double total_energy;
  unsigned int blocksize;
  int bytes_read, i, fd;
  struct stat statbuf;
  char *filename;
  char *locale;


  locale = "en_US.UTF-8";
  blocksize = 48000;
  while((i = getopt(argc, argv, "b:vl:")) != EOF){
    switch(i){
    case 'l':
      locale = optarg;
      break;
    case 'v':
      Verbose++;
      break;
    case 'b':
      blocksize = atoi(optarg);
      break;
    }
  }
  cpu_id();
  if(setlocale(LC_ALL, locale) == NULL){
    fprintf(stderr, "setlocale(%s) failed\n", locale);
  }

  if(optind < argc){
    filename = argv[optind];
    if(strcmp(filename,"-") == 0)
      filename = "/dev/stdin";

    if((fd = open(filename,O_RDONLY)) == -1){
      fprintf(stderr, "%s: can't read %s: %s\n", argv[0], argv[1], strerror(errno));
      exit(1);
    }
  } else {
    filename = NULL; // No file name argument, default to stdin
    fd = 0;
  }
  if(fstat(fd, &statbuf) != 0){
    fprintf(stderr,"%s: can't stat %s: %s\n", argv[0], filename ? filename : "[stdin]", strerror(errno));
    exit(1);
  }

  // If input is an ordinary file and not too large, try to mmap it
  // We're limited to 2^32 samples by the 'cnt' argument to sumsq()
  if((statbuf.st_mode & S_IFMT) == S_IFREG){
    samples = statbuf.st_size / sizeof(signed short);
    if(samples < UINT_MAX){
      // Input is an ordinary file and not too large; try to mmap it
      if((data = mmap(&data, statbuf.st_size, PROT_READ, MAP_FILE|MAP_SHARED, fd, 0)) != MAP_FAILED){
	// Successful mapping, read directly from memory
	ssq = sumsq(data,(int)samples);
	munmap(data,statbuf.st_size);
	printf("Samples: %'lu; RMS amplitude: %.1Lf\n", samples, sqrtl((double)ssq/samples));
	exit(0);
      }
    } else {
      fprintf(stderr,"File is %'lu bytes, too large for direct mapping\n", (unsigned long)statbuf.st_size);
    }
  }

  fprintf(stderr,"blocksize = %'u samples\n", blocksize);
  if((data = malloc(blocksize * sizeof(signed short))) == NULL){
    fprintf(stderr,"%s: can't allocate streaming buffer\n", argv[0]);
    exit(1);
  }
  total_energy = 0;
  total_samples = 0;
  for(blkno = 0;;blkno++){
    
    bytes_read = read(fd, data, blocksize * sizeof(signed short));
    if(bytes_read <= 0)
      break;
    samples = bytes_read/sizeof(signed short);
    ssq = sumsq(data,(int)samples);
    printf("block %'8llu; RMS amplitude %.1Lf\r", blkno, sqrtl((long double)ssq/samples));
    total_energy += ssq;
    total_samples += samples;
  }    
  if(bytes_read < 0)
    fprintf(stderr,"\n%s: read error: %s\n", argv[0], strerror(errno));
    
  printf("\nSamples: %'llu; RMS amplitude: %.1Lf\n", total_samples, sqrtl(total_energy/total_samples));
  exit(0);
}
