#include <limits.h>
#include "dspfec.h"

void char2short_p(signed short *,signed char *,int);
void short2int_p(signed int *,signed short *,int);
void int2ll_p(signed long long *,signed int *,int);

void char2short_sse2(signed short *,signed char *,int);
void short2int_sse2(signed int *,signed short *,int);
void int2ll_sse2(signed long long *,signed int *,int);

void short2char_p(signed char *,signed short *,int);
void int2short_p(signed short *,signed int *,int);
void ll2int_p(signed int *,signed long long *,int);

void short2char_sse2(signed char *,signed short *,int);
void int2short_sse2(signed short *,signed int *,int);
void ll2int_sse2(signed int *,signed long long *,int);

// Begin external hooks - switch to appropriate internal routine
// Expand 8-bit to 16-bit
void char2short(signed short *out,signed char *in,int cnt){
#if __i386__ || __x86_64__
  if(Cpu_features >= SSE2)
    char2short_sse2(out,in,cnt);
  else
#endif
    char2short_p(out,in,cnt);
}
// Pack signed 16-bit to 8 bit
void short2char(signed char *out,signed short *in,int cnt){
#if __i386__ || __x86_64__
  if(Cpu_features >= SSE2)
    short2char_sse2(out,in,cnt);
  else
#endif
    short2char_p(out,in,cnt);
}

// Expand signed 16-bit to 32-bit
void short2int(signed int *out,signed short *in,int cnt){
#if __i386__ || __x86_64__
  if(Cpu_features >= SSE2)
    short2int_sse2(out,in,cnt);
  else
#endif
    short2int_p(out,in,cnt);
}
// Pack signed 32-bit to 16 bit
void int2short(signed short *out,signed int *in,int cnt){
#if __i386__ || __x86_64__
  if(Cpu_features >= SSE2)
    int2short_sse2(out,in,cnt);
  else
#endif
    int2short_p(out,in,cnt);
}
// Expand signed 32 bit to 64 bit
void int2ll(signed long long *out,signed int *in,int cnt){
#if __i386__ || __x86_64__
  if(Cpu_features >= SSE2)
    int2ll_sse2(out,in,cnt);
  else
#endif
    int2ll_p(out,in,cnt);
}
// Pack signed 64 bit to signed 32 bit
void ll2int(signed int *out,signed long long *in,int cnt){
#if __i386__ || __x86_64__
  if(Cpu_features >= SSE2)
    ll2int_sse2(out,in,cnt);
  else
#endif
    ll2int_p(out,in,cnt);
}

// End of external hooks
// Beginning of portable routines

// Expand signed 8-bit to 16 bit
void char2short_p(signed short *out,signed char *in,int cnt){
  int i;

  for(i=0;i<cnt;i++)
    out[i] = in[i];
}
// Pack signed 16-bit to 8 bit
void short2char_p(signed char *out,signed short *in,int cnt){
  int i;

  for(i=0;i<cnt;i++)
    out[i] = in[i] > CHAR_MAX ? CHAR_MAX : in[i] < CHAR_MIN ? CHAR_MIN : in[i];
}

// Expand signed 16-bit to 32-bit
void short2int_p(signed int *out,signed short *in,int cnt){
  int i;

  for(i=0;i<cnt;i++)
    out[i] = in[i];
}
// Pack signed 32-bit to 16 bit
void int2short_p(signed short *out,signed int *in,int cnt){
  int i;

  for(i=0;i<cnt;i++)
    out[i] = in[i] > SHRT_MAX ? SHRT_MAX : in[i] < SHRT_MIN ? SHRT_MIN : in[i];

}
// Expand signed 32 bit to 64 bit
void int2ll_p(signed long long *out,signed int *in,int cnt){
  int i;

  for(i=0;i<cnt;i++)
    out[i] = in[i];
}
// Pack signed 64 bit to signed 32 bit
void ll2int_p(signed int *out,signed long long *in,int cnt){
  int i;

  for(i=0;i<cnt;i++)
    out[i] = in[i] > INT_MAX ? INT_MAX : in[i] < INT_MIN ? INT_MIN : in[i];

}
// End of portable routines

