#if !defined(__SSE2__)
#error "Must compile with -msse2 or better"
#endif

#include "sse.h"
int peakval_sse2(signed short *b,int cnt){
  int peak = 0,i;

  v8hi plus_peaks,minus_peaks;
  const int size_ratio = sizeof(v8hi)/sizeof(short); /* Number of shorts in a word (8) */

  if(cnt >= size_ratio){
    v8hi *buffer;
    int scnt;

    /* It's easiest to find both the negative and positive peaks, and then see which is greatest
     * when we're done. There's a packed integer absolute value instruction, but it requires SSS3.
     *
     * Do an unaligned load from the start of the buffer
     */
    plus_peaks = _mm_loadu_si128((__m128i *)b);
    minus_peaks = plus_peaks;

    /* Continue at the next 16-byte boundary
     * the overlap with the first buffer won't hurt
     */
    buffer = (v8hi *)(((long)b + 0xf) & ~0xf);
    cnt -= (short *)buffer - b;
    scnt = cnt / size_ratio;
    for(i=0;i<=scnt;i++){
      plus_peaks = _mm_max_epi16(plus_peaks,buffer[i]);
      minus_peaks = _mm_min_epi16(minus_peaks,buffer[i]);
    }
    b = (short *)&buffer[i];
    cnt -= i * size_ratio;

    /* negate the negative peaks and pick the largest */
    plus_peaks = _mm_max_epi16(plus_peaks,-minus_peaks);
    plus_peaks = _mm_max_epi16(plus_peaks,(v8hi)_mm_srli_si128(plus_peaks,8));
    plus_peaks = _mm_max_epi16(plus_peaks,(v8hi)_mm_srli_si128(plus_peaks,4));
    plus_peaks = _mm_max_epi16(plus_peaks,(v8hi)_mm_srli_si128(plus_peaks,2));
    /* The peak value from the vector search is in the low order word of plus_peaks */
    peak = _mm_extract_epi16(plus_peaks,0);
  }
  /* Now examine tail -- or whole thing if no SSE2 */
  for(i=0;i<cnt;i++){
    if(abs(b[i]) > peak)
      peak = abs(b[i]);
  }
  return peak;
}

