/* Simulation support routines
 * Add noise, etc.
 *
 * $Id$
 */
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "dspfec.h"

#define	MAX_RANDOM	0x7fffffff

int zig_init;

inline double uniform_rand(void){
  return (double)random() / MAX_RANDOM;
}


#if 1
double normal_rand(void){
  time_t t;

  if(!zig_init){
    time(&t);
    zigset(t);
    zig_init++;
  }
  return rnor();
}

#elif 0
/* Generate gaussian random double with polar rejection algorithm */
double normal_rand(){
  double fac,rsq,v1,v2;
  static double gset;
  static int iset;

  if(iset){
    /* Already got one */
    iset = 0;
    return gset;
  }
  /* Generate two evenly distributed numbers between -1 and +1
   * that are inside the unit circle
   */
  do {
    v1 = 2.0 * uniform_rand() - 1;
    v2 = 2.0 * uniform_rand() - 1;
    rsq = v1*v1 + v2*v2;
  } while(rsq >= 1.0 || rsq == 0.0);
  fac = sqrt(-2.0*log(rsq)/rsq);
  gset = v1*fac;
  iset++;
  return v2*fac;
}
#elif 0
/* Ratio-of-Uniforms */
double normal_rand(){
  const double k1 = sqrt(2/M_E);
  const double k2 = 4*exp(0.25);
  const double k3 = 4*exp(-1.35);

  double u,x,v;
  
  while(1){
    do {
      u = (double)random() / MAX_RANDOM;
    } while (u == 0.);
    v = k1 * 2.0 * (double)random() / MAX_RANDOM - 1;
    
    x = v /u;
    if(x*x <= 5 - k2*u)
      return x;
    /* First test is quicker than the second */
    if(x*x < k3/u + 1.4 && v*v < -4*u*u*log(u))
      return x;
  }
}

#else
double normal_rand(){
  const double k1 = sqrt(2/M_E);
  const double s = 0.449871;
  const double a = 0.19600;
  const double t = -0.386595;
  const double b = 0.25472;
  const double r1 = 0.27597;
  const double r2 = 0.27846;
  double u,v,x,y,Q;

  while(1){
    do {
      u = (double)random() / MAX_RANDOM;
    } while (u == 0.);
    v = k1 * 2.0 * (double)random() / MAX_RANDOM - 1;
    x = u - s;
    y = fabs(v) - t;
    Q = x*x + y*(a*y - b*x);
    if(Q < r1)
      return v/u;

    if(Q < r2 && v*v < -4*u*u*log(u))
	return v/u;
  }
}
#endif

void addnoise(unsigned char *symbols,int length,int ampl,double noise){
  float r;
  unsigned char c;

  while(length-- > 0){
    r = normal_rand();
    r = noise * r + 127.5 + (*symbols ? ampl : -ampl);

    c = r > 255 ? 255 : (r < 0 ? 0 : r);
    *symbols++ = c;
  }
}

