/* Sum of squares of 16-bit signed integers - useful for energy measurement
 * SSE2 version
 * Copyright 2009 Phil Karn, KA9Q
 * May be used under the terms of the GNU Public License (GPL)
 *
 * $Id$
 */
#include "sse.h"

#if !defined(__SSE2__)
#error "Must compile with -msse2 or better"
#endif

unsigned long long sumsq_sse2(signed short *in,int cnt){
  unsigned long long s1;
  v2di sum;
  v4si zeroes;
  v8hi *buffer;
  const int size_ratio = sizeof(v8hi)/sizeof(short); /* Number of shorts in a word (8) */
  int scnt;
  int i;

  s1 = 0;
  /* Sum unaligned words at front of buffer */
  scnt = (-(long)in & 0xf) >> 1;
  if(scnt > cnt)
    scnt = cnt;
  for(i=0; i<scnt; i++)
    s1 += (long)in[i] * in[i];

  cnt -= scnt;
  in += scnt;

  /* Bulk of work with vector instructions */
  sum = _mm_setzero_si128(); /* sum = [0] [0] */
  zeroes = (v4si) _mm_setzero_si128(); /* zeroes = [0] [0] [0] [0] */
  buffer = (v8hi *)in;
  scnt = cnt / size_ratio;
  for(i=0;i<scnt;i++){
    v4si squares;

    squares = (v4si) _mm_madd_epi16(buffer[i],buffer[i]); /* [S6*S6+S7*S7] [S4*S4+S5*S5] [S2*S2+S3*S3] [S0*S0+S1*S1] */
    /* Unpack the 32-bit partial sums into 64 bits and add to two running 64-bit sums.
     * For a general dotproduct these have to be sign-extended expands, but
     * because squares of real numbers are always positive, we can use a
     * simple interleave unpack with zero extension.
     */
    sum = _mm_add_epi64(sum,_mm_unpacklo_epi32(squares,zeroes));     /* sum += [S2*S2+S3*S3] [S0*S0+S1*S1] */ 
    sum = _mm_add_epi64(sum,_mm_unpackhi_epi32(squares,zeroes));     /* sum += [S6*S6+S7*S7] [S4*S4+S5*S5] */
  }
  sum = _mm_add_epi64(sum,_mm_unpackhi_epi64(sum,(v2di)zeroes)); /* Add upper and lower quadwords */
  cnt -= scnt * size_ratio;
  in += scnt * size_ratio;
  
  /* Mop up end of buffer */
  for(i=0;i<cnt;i++)
    s1 += (long)in[i] * in[i];

  {
    /* Vector sum is in low quadword of sum; extract and add unaligned sums */
    union {
      v2di v;
      long long a[2];
    } s;
    s.v = sum;
    s1 += s.a[0];
  }
  return s1;
}
